/**
 * X-macro list of all the monster race flags
 */

/* RF(<tier>, <index>, <name>) */

RF(1,  0, UNIQUE)
RF(1,  1, QUESTOR)
RF(1,  2, MALE)
RF(1,  3, FEMALE)
RF(1,  4, CHAR_CLEAR)
RF(1,  5, CHAR_MULTI)
RF(1,  6, ATTR_CLEAR)
RF(1,  7, ATTR_MULTI)
RF(1,  8, FORCE_DEPTH)
RF(1,  9, FORCE_MAXHP)
RF(1, 10, FORCE_SLEEP)
RF(1, 11, FORCE_EXTRA)
RF(1, 12, FRIEND)
RF(1, 13, FRIENDS)
RF(1, 14, ESCORT)
RF(1, 15, ESCORTS)
RF(1, 16, NEVER_BLOW)
RF(1, 17, NEVER_MOVE)
RF(1, 18, RAND_25)
RF(1, 19, RAND_50)
RF(1, 20, ONLY_GOLD)
RF(1, 21, ONLY_ITEM)
RF(1, 22, DROP_60)
RF(1, 23, DROP_90)
RF(1, 24, DROP_1D2)
RF(1, 25, DROP_2D2)
RF(1, 26, DROP_3D2)
RF(1, 27, DROP_4D2)
RF(1, 28, DROP_GOOD)
RF(1, 29, DROP_GREAT)
RF(1, 30, DROP_USEFUL)
RF(1, 31, DROP_CHOSEN)

RF(2,  0, STUPID)
RF(2,  1, SMART)
RF(2,  2, CAN_SPEAK)
RF(2,  3, REFLECTING)
RF(2,  4, INVISIBLE)
RF(2,  5, COLD_BLOOD)
RF(2,  6, EMPTY_MIND)
RF(2,  7, WEIRD_MIND)
RF(2,  8, DEATH_ORB)
RF(2,  9, REGENERATE)
RF(2, 10, SHAPECHANGER)
RF(2, 11, ATTR_ANY)
RF(2, 12, POWERFUL)
RF(2, 13, ELDRITCH_HORROR)
RF(2, 14, AURA_FIRE)
RF(2, 15, AURA_ELEC)
RF(2, 16, OPEN_DOOR)
RF(2, 17, BASH_DOOR)
RF(2, 18, PASS_WALL)
RF(2, 19, KILL_WALL)
RF(2, 20, MOVE_BODY)
RF(2, 21, KILL_BODY)
RF(2, 22, TAKE_ITEM)
RF(2, 23, KILL_ITEM)
RF(2, 24, BRAIN_1)
RF(2, 25, BRAIN_2)
RF(2, 26, BRAIN_3)
RF(2, 27, BRAIN_4)
RF(2, 28, BRAIN_5)
RF(2, 29, BRAIN_6)
RF(2, 30, BRAIN_7)
RF(2, 31, BRAIN_8)

RF(3,  0, ORC)
RF(3,  1, TROLL)
RF(3,  2, GIANT)
RF(3,  3, DRAGON)
RF(3,  4, DEMON)
RF(3,  5, UNDEAD)
RF(3,  6, EVIL)
RF(3,  7, ANIMAL)
RF(3,  8, THUNDERLORD)
RF(3,  9, GOOD)
RF(3, 10, AURA_COLD)
RF(3, 11, NONLIVING)
RF(3, 12, HURT_LITE)
RF(3, 13, HURT_ROCK)
RF(3, 14, SUSCEP_FIRE)
RF(3, 15, SUSCEP_COLD)
RF(3, 16, IM_ACID)
RF(3, 17, IM_ELEC)
RF(3, 18, IM_FIRE)
RF(3, 19, IM_COLD)
RF(3, 20, IM_POIS)
RF(3, 21, RES_TELE)
RF(3, 22, RES_NETH)
RF(3, 23, RES_WATE)
RF(3, 24, RES_PLAS)
RF(3, 25, RES_NEXU)
RF(3, 26, RES_DISE)
RF(3, 28, NO_FEAR)
RF(3, 29, NO_STUN)
RF(3, 30, NO_CONF)
RF(3, 31, NO_SLEEP)

RF(4,  0, AQUATIC)
RF(4,  1, CAN_SWIM)
RF(4,  2, CAN_FLY)
RF(4,  3, FRIENDLY)
RF(4,  4, PET)
RF(4,  5, MORTAL)
RF(4,  6, SPIDER)
RF(4,  7, NAZGUL)
RF(4,  8, DG_CURSE)
RF(4,  9, POSSESSOR)
RF(4, 10, NO_DEATH)
RF(4, 11, NO_TARGET)
RF(4, 12, AI_ANNOY)
RF(4, 13, AI_SPECIAL)
RF(4, 14, NEUTRAL)
RF(4, 16, DROP_RANDART)
RF(4, 17, AI_PLAYER)
RF(4, 18, NO_THEFT)
RF(4, 19, SPIRIT)

RF(5,  0, WILD_ONLY)
RF(5,  1, WILD_TOWN)
RF(5,  3, WILD_SHORE)
RF(5,  4, WILD_OCEAN)
RF(5,  5, WILD_WASTE)
RF(5,  6, WILD_WOOD)
RF(5,  7, WILD_VOLCANO)
RF(5,  9, WILD_MOUNTAIN)
RF(5, 10, WILD_GRASS)
RF(5, 11, NO_CUT)
RF(5, 15, JOKEANGBAND)
RF(5, 31, WILD_TOO)

RF(6,  0, DROP_CORPSE)
RF(6,  1, DROP_SKELETON)
RF(6,  2, HAS_LITE)
RF(6,  3, MIMIC)
RF(6,  4, HAS_EGG)
RF(6,  5, IMPRESED)
RF(6,  6, SUSCEP_ACID)
RF(6,  7, SUSCEP_ELEC)
RF(6,  8, SUSCEP_POIS)
RF(6,  9, KILL_TREES)
RF(6, 10, WYRM_PROTECT)
RF(6, 11, DOPPLEGANGER)
RF(6, 12, ONLY_DEPTH)
RF(6, 13, SPECIAL_GENE)
RF(6, 14, NEVER_GENE)
