#pragma once

/*
 * Skill constants
 */
#define SKILL_CONVEYANCE        1
#define SKILL_MANA              2
#define SKILL_FIRE              3
#define SKILL_AIR               4
#define SKILL_WATER             5
#define SKILL_NATURE            6
#define SKILL_EARTH             7
#define SKILL_SYMBIOTIC         8
#define SKILL_MUSIC             9
#define SKILL_DIVINATION        10
#define SKILL_TEMPORAL          11
#define SKILL_DRUID             12
#define SKILL_DAEMON            13
#define SKILL_META              14
#define SKILL_MAGIC             15
#define SKILL_COMBAT            16
#define SKILL_MASTERY           17
#define SKILL_SWORD             18
#define SKILL_AXE               19
#define SKILL_POLEARM           20
#define SKILL_HAFTED            21
#define SKILL_BACKSTAB          22
#define SKILL_ARCHERY           23
#define SKILL_SLING             24
#define SKILL_BOW               25
#define SKILL_XBOW              26
#define SKILL_BOOMERANG         27
#define SKILL_SPIRITUALITY      28
#define SKILL_MINDCRAFT         29
#define SKILL_MISC              30
#define SKILL_NECROMANCY        31
#define SKILL_MIMICRY           32
#define SKILL_ANTIMAGIC         33
/* XXX */
#define SKILL_STEALTH           36
/* XXX */
#define SKILL_STEALING          40
#define SKILL_SORCERY           41
#define SKILL_HAND              42
#define SKILL_THAUMATURGY       43
#define SKILL_SUMMON            44
#define SKILL_SPELL             45
#define SKILL_DODGE             46
#define SKILL_BEAR              47
#define SKILL_LORE              48
#define SKILL_PRESERVATION      49
#define SKILL_POSSESSION        50
#define SKILL_MIND              51
#define SKILL_CRITS             52
#define SKILL_PRAY              53
#define SKILL_LEARN             54
#define SKILL_UDUN              55
#define SKILL_DEVICE            56
#define SKILL_STUN              57
#define SKILL_BOULDER           58
#define SKILL_GEOMANCY          59
