/**
 * X-macro list of all the skill flags
 */

/* SKF(<tier>, <index>, <name>) */

SKF(1,  0, HIDDEN     )               /* Starts hidden */
SKF(1,  1, AUTO_HIDE  )               /* Tries to rehide at calc_bonus */
SKF(1,  2, RANDOM_GAIN)               /* Can be gained through Lost Sword quest */
