#pragma once

void do_cmd_power();

/*
 * Powers (mutation, activations, ...)
 */
#define POWER_MAX                      65

#define PWR_SPIT_ACID                  0
#define PWR_BR_FIRE                    1
#define PWR_HYPN_GAZE                  2
#define PWR_TELEKINES                  3
#define PWR_VTELEPORT                  4
#define PWR_MIND_BLST                  5
#define PWR_RADIATION                  6
#define PWR_VAMPIRISM                  7
#define PWR_SMELL_MET                  8
#define PWR_SMELL_MON                  9
#define PWR_BLINK                      10
#define PWR_EAT_ROCK                   11
#define PWR_SWAP_POS                   12
#define PWR_SHRIEK                     13
#define PWR_ILLUMINE                   14
#define PWR_BERSERK                    16
#define PWR_POLYMORPH                  17
#define PWR_MIDAS_TCH                  18
#define PWR_GROW_MOLD                  19
#define PWR_RESIST                     20
#define PWR_EARTHQUAKE                 21
#define PWR_EAT_MAGIC                  22
#define PWR_WEIGH_MAG                  23
#define PWR_STERILITY                  24
#define PWR_PANIC_HIT                  25
#define PWR_DAZZLE                     26
#define PWR_DARKRAY                    27
#define PWR_RECALL                     28
#define PWR_BANISH                     29
#define PWR_COLD_TOUCH                 30
#define PWR_LAUNCHER                   31

#define PWR_PASSWALL                    32
#define PWR_DETECT_TD                   33
#define PWR_COOK_FOOD                   34
#define PWR_UNFEAR                      35
#define PWR_EXPL_RUNE                   36
#define PWR_STM                         37
#define PWR_POIS_DART                   38
#define PWR_MAGIC_MISSILE               39
#define PWR_GROW_TREE                   40
#define PWR_BR_COLD                     41
#define PWR_BR_CHAOS                    42
#define PWR_BR_ELEM                     43
#define PWR_WRECK_WORLD                 44
#define PWR_SCARE                       45
#define PWR_REST_LIFE                   46
#define PWR_SUMMON_MONSTER              47
#define PWR_NECRO                       48
#define PWR_ROHAN                       49
#define PWR_THUNDER                     50
#define PWR_DEATHMOLD                   51
#define PWR_HYPNO                       52
#define PWR_UNHYPNO                     53
#define PWR_INCARNATE                   54
#define PWR_MAGIC_MAP                   55
#define PWR_COMPANION                   58
#define PWR_BEAR                        59
#define PWR_DODGE                       60
#define PWR_BALROG                      61
#define POWER_INVISIBILITY              62
#define POWER_WEB                       63
#define POWER_COR_SPACE_TIME            64
