#ifndef H_787198a1_aa3e_45c9_bc9f_fd7f490db78c
#define H_787198a1_aa3e_45c9_bc9f_fd7f490db78c

#include "tome/squelch/condition_metadata_fwd.hpp"

namespace squelch {

std::shared_ptr<Condition> new_condition_interactive();

} // namespace

#endif
