/**
 * X-macro list of all the ego flags
 */

/* ETR(<tier>, <index>, <name>) */

ETR(1,  0, SUSTAIN    )             /* Ego-Item gives a Random Sustain */
ETR(1,  1, OLD_RESIST )             /* The old "extra power" random high resist */
ETR(1,  2, ABILITY    )             /* Ego-Item has a random Sustain */
ETR(1,  3, R_ELEM     )             /* Item resists Acid/Fire/Cold/Elec or Poison */
ETR(1,  4, R_LOW      )             /* Item has a random low resist */
ETR(1,  5, R_HIGH     )             /* Item has a random high resist */
ETR(1,  6, R_ANY      )             /* Item has one additional resist */
ETR(1,  7, R_DRAGON   )             /* Item gets "Dragon" Resist */
ETR(1,  8, SLAY_WEAP  )             /* Special 'Slaying' bonus */
ETR(1,  9, DAM_DIE    )             /* Item has an additional dam die */
ETR(1, 10, DAM_SIZE   )             /* Item has greater damage dice */
ETR(1, 11, PVAL_M1    )             /* Item has +1 to pval */
ETR(1, 12, PVAL_M2    )             /* Item has +(up to 2) to pval */
ETR(1, 13, PVAL_M3    )             /* Item has +(up to 3) to pval */
ETR(1, 14, PVAL_M5    )             /* Item has +(up to 5) to pval */
ETR(1, 15, AC_M1      )             /* Item has +1 to AC */
ETR(1, 16, AC_M2      )             /* Item has +(up to 2) to AC */
ETR(1, 17, AC_M3      )             /* Item has +(up to 3) to AC */
ETR(1, 18, AC_M5      )             /* Item has +(up to 5) to AC */
ETR(1, 19, TH_M1      )             /* Item has +1 to hit */
ETR(1, 20, TH_M2      )             /* Item has +(up to 2) to hit */
ETR(1, 21, TH_M3      )             /* Item has +(up to 3) to hit */
ETR(1, 22, TH_M5      )             /* Item has +(up to 5) to hit */
ETR(1, 23, TD_M1      )             /* Item has +1 to dam */
ETR(1, 24, TD_M2      )             /* Item has +(up to 2) to dam */
ETR(1, 25, TD_M3      )             /* Item has +(up to 3) to dam */
ETR(1, 26, TD_M5      )             /* Item has +(up to 5) to dam */
ETR(1, 27, R_P_ABILITY)             /* Item has a random pval-affected ability */
ETR(1, 28, R_STAT     )             /* Item affects a random stat */
ETR(1, 29, R_STAT_SUST)             /* Item affects a random stat & sustains it */
ETR(1, 30, R_IMMUNITY )             /* Item gives a random immunity */
ETR(1, 31, LIMIT_BLOWS)             /* switch the "limit blows" feature */
