/**
 * X-macro list of all the dungeon flags
 */

/* DF(<tier>, <index>, <name>) */
DF(1,  0, PRINCIPAL)
DF(1,  1, MAZE)
DF(1,  2, SMALLEST)
DF(1,  3, SMALL)
DF(1,  4, BIG)
DF(1,  5, NO_DOORS)
DF(1,  6, WATER_RIVER)
DF(1,  7, LAVA_RIVER)
DF(1,  8, WATER_RIVERS)
DF(1,  9, LAVA_RIVERS)
DF(1, 10, CAVE)
DF(1, 11, CAVERN)
DF(1, 12, NO_UP)
DF(1, 13, HOT)
DF(1, 14, COLD)
DF(1, 15, FORCE_DOWN)
DF(1, 16, FORGET)
DF(1, 17, NO_DESTROY)
DF(1, 18, SAND_VEIN)
DF(1, 19, CIRCULAR_ROOMS)
DF(1, 20, EMPTY)
DF(1, 21, DAMAGE_FEAT)
DF(1, 22, FLAT)
DF(1, 23, TOWER)
DF(1, 24, RANDOM_TOWNS)
DF(1, 25, DOUBLE)
DF(1, 26, LIFE_LEVEL)
DF(1, 27, EVOLVE)
DF(1, 28, ADJUST_LEVEL_1)
DF(1, 29, ADJUST_LEVEL_2)
DF(1, 30, NO_RECALL)
DF(1, 31, NO_STREAMERS)

DF(2,  0, ADJUST_LEVEL_1_2)
DF(2,  1, NO_SHAFT)
DF(2,  2, ADJUST_LEVEL_PLAYER)
DF(2,  3, NO_TELEPORT)
DF(2,  4, ASK_LEAVE)
DF(2,  5, NO_STAIR)
DF(2,  6, SPECIAL)
DF(2,  7, NO_NEW_MONSTER)
DF(2,  9, NO_GENO)
DF(2, 10, NO_BREATH)
DF(2, 11, WATER_BREATH)
DF(2, 12, ELVEN)
DF(2, 13, DWARVEN)
DF(2, 14, NO_EASY_MOVE)
DF(2, 15, NO_RECALL_OUT)
