#ifndef H_7922f591_bdfd_491d_8561_b11225285fea
#define H_7922f591_bdfd_491d_8561_b11225285fea

#include "tome/squelch/condition_fwd.hpp"

#include <memory>

namespace squelch {

std::shared_ptr<Condition> new_condition_interactive();

} // namespace

#endif
