/**
 * X-macro list of all the player race flags
 */

/* PR(<tier>, <index>, <name>) */

PR(1,  0, EXPERIMENTAL)                   /* Experimental */
PR(1,  1, RESIST_BLACK_BREATH)            /* Resists black breath */
PR(1,  2, NO_STUN)                        /* Cannot be stunned */
PR(1,  3, XTRA_MIGHT_BOW)                 /* Xtra might with bows */
PR(1,  4, XTRA_MIGHT_XBOW)                /* Xtra might with x-bows */
PR(1,  5, XTRA_MIGHT_SLING)               /* Xtra might with slings */
PR(1,  6, AC_LEVEL)                       /* AC increases with levels */
PR(1,  7, HURT_LITE)                      /* Hurt by light */
PR(1,  8, VAMPIRE)                        /* Vampire */
PR(1,  9, UNDEAD)                         /* Undead */
PR(1, 10, NO_CUT)                         /* No cuts */
PR(1, 11, CORRUPT)                        /* Corrupted; automatically gains corruptions */
PR(1, 12, NO_FOOD)                        /* Little gain from food */
PR(1, 13, NO_GOD)                         /* Cannot worship any gods */
PR(1, 14, ELF)                            /* Is an elf */
PR(1, 15, SEMI_WRAITH)                    /* Allows walking through walls; taking damage */
PR(1, 16, NO_SUBRACE_CHANGE)              /* Impossible to change subrace */
PR(1, 17, GOD_FRIEND)                     /* Better grace */
PR(1, 18, EASE_STEAL)                     /* Gain XP by stealing */
PR(1, 19, ASTRAL)                         /* Astral being from the halls of Mandos */
