/**
 * X-macro list of all the store flags
 */

/* STF(<tier>, <index>, <name>) */

STF(1,  0, DEPEND_LEVEL )
STF(1,  1, SHALLOW_LEVEL)
STF(1,  2, MEDIUM_LEVEL )
STF(1,  3, DEEP_LEVEL   )
STF(1,  4, RARE	       )
STF(1,  5, VERY_RARE    )
STF(1,  6, COMMON       )
STF(1,  7, ALL_ITEM     )
STF(1,  8, RANDOM       )
STF(1,  9, FORCE_LEVEL  )
STF(1, 10, MUSEUM       )
