/**
 * X-macro list of all the monster race flags
 */

/* SF(tier, index, name, is_summon, is_annoy, is_damage, is_bolt, is_smart, is_innate, is_escape, is_tactic, is_haste, is_heal) */

SF(1,  0, SHRIEK       , false, true , false, false, false, true , false, false, false, false)
SF(1,  1, MULTIPLY     , false, false, false, false, false, true , false, false, false, false)
SF(1,  2, S_ANIMAL     , true , false, false, false, true , false, false, false, false, false)
SF(1,  3, ROCKET       , false, false, true , false, false, true , false, false, false, false)
SF(1,  4, ARROW_1      , false, false, true , true , false, true , false, false, false, false)
SF(1,  5, ARROW_2      , false, false, true , true , false, true , false, false, false, false)
SF(1,  6, ARROW_3      , false, false, true , true , false, true , false, false, false, false)
SF(1,  7, ARROW_4      , false, false, true , true , false, true , false, false, false, false)
SF(1,  8, BR_ACID      , false, false, true , false, false, true , false, false, false, false)
SF(1,  9, BR_ELEC      , false, false, true , false, false, true , false, false, false, false)
SF(1, 10, BR_FIRE      , false, false, true , false, false, true , false, false, false, false)
SF(1, 11, BR_COLD      , false, false, true , false, false, true , false, false, false, false)
SF(1, 12, BR_POIS      , false, false, true , false, false, true , false, false, false, false)
SF(1, 13, BR_NETH      , false, false, true , false, false, true , false, false, false, false)
SF(1, 14, BR_LITE      , false, false, true , false, false, true , false, false, false, false)
SF(1, 15, BR_DARK      , false, false, true , false, false, true , false, false, false, false)
SF(1, 16, BR_CONF      , false, false, true , false, false, true , false, false, false, false)
SF(1, 17, BR_SOUN      , false, false, true , false, false, true , false, false, false, false)
SF(1, 18, BR_CHAO      , false, false, true , false, false, true , false, false, false, false)
SF(1, 19, BR_DISE      , false, false, true , false, false, true , false, false, false, false)
SF(1, 20, BR_NEXU      , false, false, true , false, false, true , false, false, false, false)
SF(1, 21, BR_TIME      , false, false, true , false, false, true , false, false, false, false)
SF(1, 22, BR_INER      , false, false, true , false, false, true , false, false, false, false)
SF(1, 23, BR_GRAV      , false, false, true , false, false, true , false, false, false, false)
SF(1, 24, BR_SHAR      , false, false, true , false, false, true , false, false, false, false)
SF(1, 25, BR_PLAS      , false, false, true , false, false, true , false, false, false, false)
SF(1, 26, BR_WALL      , false, false, true , false, false, true , false, false, false, false)
SF(1, 27, BR_MANA      , false, false, true , false, false, true , false, false, false, false)
SF(1, 28, BA_NUKE      , false, false, true , false, false, false, false, false, false, false)
SF(1, 29, BR_NUKE      , false, false, true , false, false, true , false, false, false, false)
SF(1, 30, BA_CHAO      , false, false, true , false, false, false, false, false, false, false)
SF(1, 31, BR_DISI      , false, false, true , false, false, true , false, false, false, false)
SF(2,  0, BA_ACID      , false, false, true , false, false, false, false, false, false, false)
SF(2,  1, BA_ELEC      , false, false, true , false, false, false, false, false, false, false)
SF(2,  2, BA_FIRE      , false, false, true , false, false, false, false, false, false, false)
SF(2,  3, BA_COLD      , false, false, true , false, false, false, false, false, false, false)
SF(2,  4, BA_POIS      , false, false, true , false, false, false, false, false, false, false)
SF(2,  5, BA_NETH      , false, false, true , false, false, false, false, false, false, false)
SF(2,  6, BA_WATE      , false, false, true , false, false, false, false, false, false, false)
SF(2,  7, BA_MANA      , false, false, true , false, false, false, false, false, false, false)
SF(2,  8, BA_DARK      , false, false, true , false, false, false, false, false, false, false)
SF(2,  9, DRAIN_MANA   , false, true , false, false, false, false, false, false, false, false)
SF(2, 10, MIND_BLAST   , false, true , false, false, false, false, false, false, false, false)
SF(2, 11, BRAIN_SMASH  , false, true , false, false, false, false, false, false, false, false)
SF(2, 12, CAUSE_1      , false, true , true , false, false, false, false, false, false, false)
SF(2, 13, CAUSE_2      , false, true , true , false, false, false, false, false, false, false)
SF(2, 14, CAUSE_3      , false, true , true , false, false, false, false, false, false, false)
SF(2, 15, CAUSE_4      , false, true , true , false, false, false, false, false, false, false)
SF(2, 16, BO_ACID      , false, false, true , true , false, false, false, false, false, false)
SF(2, 17, BO_ELEC      , false, false, true , true , false, false, false, false, false, false)
SF(2, 18, BO_FIRE      , false, false, true , true , false, false, false, false, false, false)
SF(2, 19, BO_COLD      , false, false, true , true , false, false, false, false, false, false)
SF(2, 20, BO_POIS      , false, false, true , true , false, false, false, false, false, false)
SF(2, 21, BO_NETH      , false, false, true , true , false, false, false, false, false, false)
SF(2, 22, BO_WATE      , false, false, true , true , false, false, false, false, false, false)
SF(2, 23, BO_MANA      , false, false, true , true , false, false, false, false, false, false)
SF(2, 24, BO_PLAS      , false, false, true , true , false, false, false, false, false, false)
SF(2, 25, BO_ICEE      , false, false, true , true , false, false, false, false, false, false)
SF(2, 26, MISSILE      , false, false, true , true , false, false, false, false, false, false)
SF(2, 27, SCARE        , false, true , false, false, true , false, false, false, false, false)
SF(2, 28, BLIND        , false, true , false, false, true , false, false, false, false, false)
SF(2, 29, CONF         , false, true , false, false, true , false, false, false, false, false)
SF(2, 30, SLOW         , false, true , false, false, true , false, false, false, false, false)
SF(2, 31, HOLD         , false, true , false, false, true , false, false, false, false, false)
SF(3,  0, HASTE        , false, false, false, false, true , false, false, false, true , false)
SF(3,  1, HAND_DOOM    , false, false, true , false, false, false, false, false, false, false)
SF(3,  2, HEAL         , false, false, false, false, true , false, false, false, false, true )
SF(3,  3, S_ANIMALS    , true , false, false, false, true , false, false, false, false, false)
SF(3,  4, BLINK        , false, false, false, false, true , false, true , true , false, false)
SF(3,  5, TPORT        , false, false, false, false, true , false, true , false, false, false)
SF(3,  6, TELE_TO      , false, true , false, false, false, false, false, false, false, false)
SF(3,  7, TELE_AWAY    , false, false, false, false, true , false, true , false, false, false)
SF(3,  8, TELE_LEVEL   , false, false, false, false, true , false, true , false, false, false)
SF(3,  9, DARKNESS     , false, true , false, false, false, false, false, false, false, false)
SF(3, 11, FORGET       , false, true , false, false, false, false, false, false, false, false)
SF(3, 12, RAISE_DEAD   , false, false, false, false, false, false, false, false, false, false)
SF(3, 13, S_BUG        , true , false, false, false, true , false, false, false, false, false)
SF(3, 14, S_RNG        , true , false, false, false, true , false, false, false, false, false)
SF(3, 15, S_THUNDERLORD, true , false, false, false, true , false, false, false, false, false)
SF(3, 16, S_KIN        , true , false, false, false, true , false, false, false, false, false)
SF(3, 17, S_HI_DEMON   , true , false, false, false, true , false, false, false, false, false)
SF(3, 18, S_MONSTER    , true , false, false, false, true , false, false, false, false, false)
SF(3, 19, S_MONSTERS   , true , false, false, false, true , false, false, false, false, false)
SF(3, 20, S_ANT        , true , false, false, false, true , false, false, false, false, false)
SF(3, 21, S_SPIDER     , true , false, false, false, true , false, false, false, false, false)
SF(3, 22, S_HOUND      , true , false, false, false, true , false, false, false, false, false)
SF(3, 23, S_HYDRA      , true , false, false, false, true , false, false, false, false, false)
SF(3, 24, S_ANGEL      , true , false, false, false, true , false, false, false, false, false)
SF(3, 25, S_DEMON      , true , false, false, false, true , false, false, false, false, false)
SF(3, 26, S_UNDEAD     , true , false, false, false, true , false, false, false, false, false)
SF(3, 27, S_DRAGON     , true , false, false, false, true , false, false, false, false, false)
SF(3, 28, S_HI_UNDEAD  , true , false, false, false, true , false, false, false, false, false)
SF(3, 29, S_HI_DRAGON  , true , false, false, false, true , false, false, false, false, false)
SF(3, 30, S_WRAITH     , true , false, false, false, true , false, false, false, false, false)
SF(3, 31, S_UNIQUE     , true , false, false, false, true , false, false, false, false, false)
