#pragma once

#include "h-basic.hpp"

struct power_activation {

	byte level;             /* Min level */
	byte cost;              /* Mana/Life cost */
	byte stat;              /* Stat used */
	byte diff;              /* Difficulty */

};
