
//          Copyright Joakim Karlsson & Kim Gräsman 2010-2012.
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

#ifndef IGLOO_CONSTRAINTS_H
#define IGLOO_CONSTRAINTS_H

#include "containsconstraint.h"
#include "endswithconstraint.h"
#include "equalsconstraint.h"
#include "haslengthconstraint.h"
#include "isgreaterthanconstraint.h"
#include "isgreaterthanorequaltoconstraint.h"
#include "islessthanconstraint.h"
#include "islessthanorequaltoconstraint.h"
#include "startswithconstraint.h"
#include "fulfillsconstraint.h"
#include "equalswithdeltaconstraint.h"
#include "equalscontainerconstraint.h"

#endif
