#pragma once

void do_cmd_store();
void store_shuffle(int which);
void store_maint(int town_num, int store_num);
void store_init(int town_num, int store_num);
void do_cmd_home_trump();
void store_sell();
void store_purchase();
void store_examine();
void store_stole();
void store_prt_gold();
