#pragma once

#include "h-basic.hpp"

void do_cmd_html_dump();
void cli_add(const char *active, const char *trigger, const char *descr);
void do_cmd_cli();
void do_cmd_cli_help();
void do_cmd_inven();
void do_cmd_equip();
void do_cmd_wield();
void do_cmd_takeoff();
void do_cmd_drop();
void do_cmd_destroy();
void do_cmd_observe();
void do_cmd_uninscribe();
void do_cmd_inscribe();
void do_cmd_refill();
void do_cmd_target();
void do_cmd_look();
void do_cmd_locate();
void do_cmd_query_symbol();
void do_cmd_sense_grid_mana();
